// Example from README file.
// Demonstrates the Woomera Dart package.

import 'dart:io';
import 'dart:async';
import 'package:woomera/woomera.dart';
import 'package:logging/logging.dart';

//================================================================

Future main() async {
  loggingSetup();

  // Create and configure Web server

  var ws = new Server();
  ws.bindAddress = InternetAddress.ANY_IP_V6;
  ws.bindPort = 1024;
  ws.exceptionHandler = myExceptionHandler;

  var p = ws.pipelines.first; // get the first/default pipeline

  // Register rules in the pipeline. The "~/" indicates the path is relative
  // to the deployment path (which is usually "/", but does not have to be).

  p.get("~/", handleTopLevel);
  p.get("~/calculate/:units/:month/:day", handleCalculate);
  p.post("~/transaction", handleTransaction);
  p.get("~/login", handleLoginForm);
  p.post("~/login", handleLoginRequest);
  p.get("~/logout", handleLogoutForm);
  p.post("~/logout", handleLogoutRequest);

  var projectDir = FileSystemEntity
      .parentOf(FileSystemEntity.parentOf(Platform.script.path));
  var cssDir = projectDir + "/web/style";

  p.get("~/css/*", new StaticFiles(cssDir).handler);

  // Run Web server

  await ws.run();
}

//----------------------------------------------------------------

void loggingSetup() {
  hierarchicalLoggingEnabled = true;
  Logger.root.onRecord.listen((LogRecord rec) {
    print('${rec.time}: ${rec.loggerName}: ${rec.level.name}: ${rec.message}');
  });

  Logger.root.level = Level.OFF;

  var commonLevel = Level.INFO;

  new Logger("main").level = commonLevel;
  new Logger("woomera.server").level = commonLevel;
  new Logger("woomera.request").level = Level.FINE;
  new Logger("woomera.request.header").level = commonLevel;
  new Logger("woomera.request.param").level = commonLevel;
  new Logger("woomera.response").level = commonLevel;
  new Logger("woomera.session").level = commonLevel;
}

//----------------------------------------------------------------

Future<Response> handleTopLevel(Request req) async {
  var username = (req.session != null) ? req.session["username"] : null;

  var resp = new ResponseBuffered(ContentType.HTML);

  // Generate the response. All URLs are generated by calling req.rewriteUrl
  // to ensure they include the deployment path and the session is preserved
  // using URL rewriting (if session cookies cannot be used).

  var authLink = (username == null) ? "~/login" : "~/logout";
  var authText = (username == null) ? "Login" : "Logout";

  resp.write("""
<html>
  <head>
    <title>Woomera Example</title>
    <link rel="stylesheet" href="${req.rewriteUrl("~/css/site.css")}"/>
  </head>
  <body>
    <h1>Welcome ${(username != null) ? username : ""}</h1>
    <p class="login-logout"><a href="${req.rewriteUrl(authLink)}">$authText</a></p>

    <p>Days to <a href="${req.rewriteUrl("~/calculate/days/12/25")}">Christmas</a>
       and <a href="${req.rewriteUrl("~/calculate/days/1/1")}">New Year</a>;
       weeks to <a href="${req.rewriteUrl("~/calculate/weeks/12/25")}">Christmas</a>
       and <a href="${req.rewriteUrl("~/calculate/weeks/1/1")}">New Year</a>.</p>

    <p>
    <form method="POST" action="${req.rewriteUrl("~/transaction")}">
      <span class="balance">Balance: $balance</span>
      <input type="radio" name="action" id="d" value="deposit"/><label for="d">Deposit</label>
      <input type="radio" name="action" id="w" value="withdraw"/><label for="w">Withdraw</label>
      <input type="number" name="quantity"/>
      <input type="submit"/>
    </form>
    </p>

    <p>Examples of exception handling:
      <a href="${req.rewriteUrl("~/no-such-page")}">missing page</a>,
      <a href="${req.rewriteUrl("~/calculate/minutes/12/25")}">invalid unit</a>,
      <a href="${req.rewriteUrl("~/calculate/days/13/01")}">invalid date</a>.
    </p>
  </body>
</html>
""");
  return resp;
}

//================================================================
// GET example

Future<Response> handleCalculate(Request req) async {
  var units = req.pathParams["units"];
  var month = int.parse(req.pathParams["month"]);
  var day = int.parse(req.pathParams["day"]);

  var today = new DateTime.now();
  var targetDate = new DateTime(today.year, month, day);
  if (targetDate.isBefore(today)) {
    targetDate = new DateTime(today.year + 1, month, day);
  }
  if (targetDate.month != month) {
    throw new RangeError.value(month, "month", "Out of range");
  }
  if (targetDate.day != day) {
    throw new RangeError.value(day, "day", "Out of range");
  }

  var delay = targetDate.difference(today);
  var num;
  switch (units) {
    case "weeks":
      num = delay.inDays ~/ 7;
      break;
    case "days":
      num = delay.inDays + 1;
      break;
    default:
      throw new ArgumentError.value(units, "units", "Unexpected value");
  }

  var resp = new ResponseBuffered(ContentType.HTML);
  resp.write("""
<html>
  <head>
    <title>Date calculator</title>
    <link rel="stylesheet" href="${req.rewriteUrl("~/css/site.css")}"/>
  </head>
  <body>
    <h1>Date calculator</h1>
    <p>There are $num $units until ${targetDate.toIso8601String().substring(0, 10)}.</p>
    <p class="home"><a href="${req.rewriteUrl("~/")}">Home</a></p>
  </body>
</html>
""");
  return resp;
}

//================================================================
// POST example

int balance = 0;

Future<Response> handleTransaction(Request req) async {
  try {
    var quantity = int.parse(req.postParams["quantity"]);
    var action = req.postParams["action"];

    switch (action) {
      case "deposit":
        balance += quantity;
        break;
      case "withdraw":
        balance -= quantity;
        break;
      default:
        assert(action == "");
        throw new ArgumentError.value(action, "action", "Unknown value");
        break;
    }

    return new ResponseRedirect("~/");
  } on FormatException catch (_) {
    throw new InvalidAmount();
  }
}

//================================================================
// Sessions

const String testCookieName = "testCookie";

Future<Response> handleLoginForm(Request req) async {
  if (req.session != null) {
    throw new StateError("Cannot show login form: already logged in");
  }

  var resp = new ResponseBuffered(ContentType.HTML);

  var testCookie = new Cookie(testCookieName, "cookies_work!");
  testCookie.path = req.server.basePath;
  testCookie.httpOnly = true;
  resp.cookieAdd(testCookie);

  resp.write("""
<html>
  <head>
    <title>Login</title>
    <link rel="stylesheet" href="${req.rewriteUrl("~/css/site.css")}"/>
  </head>
  <body>
    <h1>Login</h1>
    <form method="POST" action="${req.rewriteUrl("~/login")}">
      <label for="uname">Username</label>
      <input type="text" name="username" id="uname"/>
      <input type="submit" value="Login"/>
    </form>
  </body>
</html>
""");

  return resp;
}

class UsernameMissing implements Exception {}

class InvalidAmount implements Exception {}

Future<Response> handleLoginRequest(Request req) async {
  if (req.session != null) {
    throw new StateError("Cannot process login request: already logged in");
  }

  var username = req.postParams["username"];
  if (username.isEmpty) {
    throw new UsernameMissing();
  }
  var session = new Session(req.server, new Duration(minutes: 10));
  session["username"] = username;
  session["timestamp"] = new DateTime.now();

  req.session = session;

  var resp = new ResponseRedirect("~/");
  resp.cookieDelete(testCookieName, req.server.basePath);
  return resp;
}

Future<Response> handleLogoutForm(Request req) async {
  if (req.session == null) {
    throw new StateError("Cannot show logout form: not logged in");
  }

  var resp = new ResponseBuffered(ContentType.HTML);

  resp.write("""
<html>
  <head>
    <title>Logout</title>
    <link rel="stylesheet" href="${req.rewriteUrl("~/css/site.css")}"/>
  </head>
  <body>
    <h1>Logout</h1>
    <form method="POST" action="${req.rewriteUrl("~/logout")}">
      <input type="submit" value="Logout"/>
    </form>
  </body>
</html>
""");

  return resp;
}

Future<Response> handleLogoutRequest(Request req) async {
  if (req.session == null) {
    throw new StateError("Cannot process logout request: not logged in");
  }

  await req.session.terminate();
  req.session = null;
  return new ResponseRedirect("~/");
}

//================================================================
// Exception handler

Future<Response> myExceptionHandler(
    Request req, Object ex, StackTrace st) async {
  var status = HttpStatus.INTERNAL_SERVER_ERROR;
  var message;

  if (ex is NotFoundException) {
    status = (ex.found == NotFoundException.foundNothing)
        ? HttpStatus.METHOD_NOT_ALLOWED
        : HttpStatus.NOT_FOUND;
    message = "Sorry, the page you were looking for could not be found.";
  } else if (ex is UsernameMissing) {
    message = "Username cannot be blank.";
  } else if (ex is ArgumentError) {
    status = HttpStatus.BAD_REQUEST;
    message = "Bad values provided in request.";
  } else if (ex is InvalidAmount) {
    status = HttpStatus.BAD_REQUEST;
    message = "Bad amount: must be an integer.";
  } else {
    // Hide details from user and just show a user friendly message.
    status = HttpStatus.INTERNAL_SERVER_ERROR;
    message = "Sorry, an internal error occured. It wasn't your fault.";
    new Logger("main").shout("Internal exception: $ex");
  }

  var resp = new ResponseBuffered(ContentType.HTML);
  resp.status = status;

  resp.write("""
<html>
  <head>
    <title>Error</title>
    <link rel="stylesheet" href="${req.rewriteUrl("~/style/site.css")}"/>
  </head>
  <body>
    <h1>Error</h1>
    <p>${HEsc.text(message)}</p>
    <p class="home"><a href="${req.rewriteUrl("~/")}">Home</a></p>
  </body>
</html>
""");

  return resp;
}
